/****************************************************************************\
  This file contains functions that can be called as events.
  All the functions *MUST* use ev_args as defined in schedule.h

  An event can be generated thus;

  add_event ( 10, 0, ev_saytoplayer, "ds", playerno, "Chickens!\n" )

  			                   ^^^^^^^^^ Arguments
					   (read/specific to function)
				     ^^^ format of args
                      ^^^ Function to call (void return)
               ^^^^^ Seconds and Micro seconds to call event

\****************************************************************************/
#include <sys/types.h>
#include "global.h"
#include "daemon.h"
#include "persona.h"
#include "schedule.h"		/* Includes varargs and stdio.h!!! */
#include "events.h"
#include "animation.h"


/****************************************************************************\
  This function is like form but for ev_lists....
  Just pass the ev_list and all will be formatted and return as a string
\****************************************************************************/

char *ev_form ( pvar )
    ev_list pvar;
{
    int l;
    static char buffer[LINELEN];
    buffer[0] = 0;

    do {
	switch ( ev_typeof ( pvar ) ) {
	case STRING:
	    strcat ( buffer, ev_string ( pvar ) );
	    break;
	case CHAR:
	    l = strlen ( buffer );
	    buffer[++l] = ev_char ( pvar );
	    buffer[++l] = 0;
	    break;
	default:
	    strcat ( buffer, form ( "%d", ev_int ( pvar ) ) );
	    break;
	}
    } while ( ! ev_isend ( pvar ) );
    return ( buffer );
}

/****************************************************************************\
  ev_saytoplayer takes a playerno, and messages Ie;

  add_event ( sec, usec, ev_saytoplayer, "dss", 5, "Chickens", "cheese" );

\****************************************************************************/

void ev_saytoplayer ( ev_alist )
    ev_dcl
{
    ev_list pvar;		/* Parameter traversing variable */
    int playerno;		/* Playerno to send to */

    ev_start (pvar);		/* Start pvar */

    playerno = ev_int ( pvar );	/* Get playerno */
    safewrite ( playerno, ev_form ( pvar ) );

    ev_end (pvar);		/* End pvar */
}

/****************************************************************************\
  ev_saytoroom takes a roomno, and a format message (with args). Ie;

  add_event ( sec, usec, ev_saytoroom, "dss", LIMBO, "Chickens", "cheese\n" );

\****************************************************************************/

void ev_saytoroom ( ev_alist )
    ev_dcl
{
    ev_list pvar;		/* Parameter traversing variable */
    int where;			/* Where to write message */

    ev_start (pvar);		/* Start pvar */

    where = ev_int ( pvar );	/* Get where */
    speak ( ev_form ( pvar ), NO_ONE, where );

    ev_end (pvar);		/* End pvar */
}

/****************************************************************************\
  ev_shout takes just format message (with args). Ie;

  NB. This shout will be ignored if IGNORESHOUT is set!

  add_event ( sec, usec, ev_shout, "sss", "Chickens", fish, "!\n" );

\****************************************************************************/

void ev_shout ( ev_alist )
    ev_dcl
{
    ev_list pvar;		/* Parameter traversing variable */

    ev_start (pvar);		/* Start pvar */
    shout ( ev_form ( pvar ), NO_ONE, MESG_SHOUT );
    ev_end (pvar);		/* End pvar */
}

/****************************************************************************\
  ev_warn takes just format message (with args). Ie;

  NB. This warn will be ignored if IGNOREWARN is set!

  add_event ( sec, usec, ev_warn, "sss", "Chickens", fish, "!\n" );

\****************************************************************************/

void ev_warn ( ev_alist )
    ev_dcl
{
    ev_list pvar;		/* Parameter traversing variable */

    ev_start (pvar);		/* Start pvar */
    shout ( ev_form ( pvar ), NO_ONE, MESG_WARN );
    ev_end (pvar);		/* End pvar */
}

/****************************************************************************\
  Reconnect limbo exit to start.

  Format: ""  (no arguments)
\****************************************************************************/

void ev_limboreconnect ( ev_alist )
    ev_dcl
{
    static int lastexit = 8;	/* Out */

    loc[0].exits[lastexit] = -1; /* Old exit is to limbo */
    lastexit = (rand () % (no_directions-1)) + 1; /* New exit */
    loc[0].exits[lastexit] = STARTLOC;

				/* Setup for next time ;-} */
    add_event ( NEXT_LIMBO, 0, ev_limboreconnect, "" );
}

/****************************************************************************\
  Recompute which objects can move - this should be performed every
  so often. (10 seconds)
\****************************************************************************/

void ev_computeanimate ( ev_alist )
    ev_dcl
{
    int ind;
				/* Find an animate object */
    num_anim = 0;
    for ( ind = 0; ind < no_objects; ind++ )
	if ( animatep(ind) )
	    anim_list[num_anim++] = ind;

    add_event ( COMPUTE_ANIM, 0, ev_computeanimate, "" );
}

/****************************************************************************\
  Move a cluck (animate object) every so often.... kinda witty huh!
\****************************************************************************/

void ev_moveobjects ( ev_alist )
    ev_dcl
{
    int sec, usec;		/* When is next event? */
    int ind;
				/* Must be some moveable clucks! */
    if ( num_anim > 0 ) {
				/* Pick yer cluck */
	ind = anim_list[(random () % num_anim)];

	if ( (random () % 100) <= obj[ind].speed )
	    animate_object ( ind );

				/* Next event will be in 1-10 seconds */
				/* divided by number of moveable objects */
				/* (minimum .05 second) */
	usec  = ((random () % 9) + 1) 
	    * (1000000 / num_anim);
	sec   = usec / 1000000;

	usec %= 1000000;
	if ( sec == 0 && usec < 050000 )
	    usec = 50000;

    } else {
	sec = COMPUTE_ANIM;	/* If no moving objects... */
	usec = 0;		/* Retry after COMPUTE_ANIM */

    }

    add_event ( sec, usec, ev_moveobjects, "" );
}

/****************************************************************************\
  Handle command from person.

  Form: "d...": 

  eg

  ev_handle_command, "dssss", playerno, playername,
                     "kill ", name, " with chicken\n"

\****************************************************************************/

void ev_handle_command ( ev_alist )
    ev_dcl
{
    ev_list pvar;
    int playerno;
    char *name;			/* Name of character - must be same! */

    ev_start ( pvar );

    playerno = ev_int ( pvar );	/* Get playerno */
				/* Get player name */
    name     = ev_string ( pvar );
				/* Only attempt if player is on */
    if (   (player[playerno].fd != -1)
	&& (strcmp ( name, player[playerno].name ) == 0) )
	handle ( playerno, ev_form ( pvar ) );

    ev_end ( pvar );
}

/****************************************************************************\
  Close all connections... do it nicely... Like give some warnings...
\****************************************************************************/
void ev_close_mundi ( ev_alist )
    ev_dcl
{
				/* Kill all non-Arches */
    add_event ( WARN1_TIME, 0, ev_killall_players, "" );

				/* First warning */
    shout ( form ( "*** Mundi will be closing in %d minutes ***\n",
		  WARN1_TIME ), NO_ONE, MESG_WARN );

				/* Second warning */
    add_event ( WARN2_TIME, 0,
	       ev_warn, "sis",
	       "*** Mundi will be closing in ",
	       (WARN2_TIME/60),
	       " minutes ***\n" );

				/* Third warning */
    add_event ( ( WARN2_TIME + WARN3_TIME ), 0,
	       ev_warn, "sis",
	       "*** Mundi will be closing in ",
	       (WARN3_TIME/60),
	       " minutes ***\n" );

				/* Fourth warning */
    add_event ( ( WARN2_TIME + WARN3_TIME + WARN4_TIME), 0,
	       ev_warn, "sis",
	       "*** Mundi will be closing in ",
	       WARN4_TIME,
	       " seconds ***" );


				/* Final warning */
    add_event ( ( WARN2_TIME + WARN3_TIME + WARN4_TIME + WARN5_TIME), 0,
	       ev_warn, "sis",
	       "*** Last shutdown warning: Logout or die in ",
	       WARN5_TIME,
	       " seconds ***\n" );
}

/****************************************************************************\
  Save and logout all players
\****************************************************************************/

void ev_killall_players ( ev_alist )
    ev_dcl
{
    int i, d, h = next_open_time ( &d ), s;

    for ( i = 0; i < MAXUSERS; i++ )
	if ( player[i].fd != -1 && !CANARCHP(i) ) {
	    safewrite ( i, form ( "%s%s %d:00 hours", NOTOPEN, days[d], h ) );
	    killoff ( i );
	}
    logerror ( "Mundi *Closedown* completed.\n" );

				/* Set event for opening */
    h = next_open_time ( &d );
    if ( ( s = compute_seconds_till ( d, h ) ) < 0 )
	s = 0;
    add_event ( s, 0, ev_open_mundi, "" );
}

/****************************************************************************\
  Open mundi
\****************************************************************************/

void ev_open_mundi ( ev_alist )
    ev_dcl
{
    int d, h = next_close_time ( &d ), s;
    if ( ( s = (compute_seconds_till ( d, h ) - WARN1_TIME) ) < 0 )
	s = 0;
    add_event ( s, 0, ev_close_mundi, "" );
    logerror ( "Mundi *Open* completed.\n" );
}

/****************************************************************************\
  Startup events
\****************************************************************************/
void ev_startevents ( ev_alist )
    ev_dcl
{
    add_event ( 0, 0, ev_computeanimate, "" );
    add_event ( 0, 0, ev_limboreconnect, "" );
    add_event ( 0, 0, ev_moveobjects, "" );
}

