/***************************************************************************\
|*                                                                         *|
|*  Hhelp.c:	A module to implement a heirachical VMS-like "help"        *|
|*		system using the UNIX file-system as a model.  This        *|
|*		is a set of library-like functions that should be          *|
|*		callable from any program that uses the specified          *|
|*		interface, but was developed for MUNDI.			   *|
|*                                                                         *|
|*  Started:	Sun May 21 23:51:23 BST 1989				   *|
|*                                                                         *|
\***************************************************************************/

/***************************************************************************\
|*                                                                         *|
|*  We need the definition of NULL in several places below, and it is      *|
|*  silly to include the whole of stdio.h just for that; therefore we      *|
|*  omit it by default, since dirent.h also defines it (at least, it	   *|
|*  does here at Warwick.  If your dirent.h does not have NULL defined,    *|
|*  then we include stdio.h anyway.  That's what I always say.		   *|
|*                                                                         *|
|*  It is possible that your system does not have dirent.h, which is a     *|
|*  new bit of Berkeley-ism (or maybe SunOSism, who knows?)  In this       *|
|*  case, replace its #include with one to sys/dir.h, and change all	   *|
|*  ocurrences of "struct dirent" to "struct direct".			   *|
|*                                                                         *|
\***************************************************************************/

#include <strings.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <varargs.h>
#include <fcntl.h>
#include <ctype.h>

#ifndef NULL
#  include <stdio.h>
#endif /*NULL*/

/*-------------------------------------------------------------------------*/

#define HH_LINELEN 160		/* Not in hhelp.h 'cos I have no desire */
				/* for it to be publicly visible. */

/*-------------------------------------------------------------------------*/

/*extern char *sprintf ();*/	/* 2001: This is declared in <stdio.h> */
extern char *malloc ();
extern char *realloc ();
extern char *strdup ();

/*-------------------------------------------------------------------------*/

static char hh_root[HH_LINELEN];
static char hh_current[HH_LINELEN];
static char **hh_files;
static int hh_nofiles;

/***************************************************************************\
|*                                                                         *|
|*  The function hh_samestring() isn't really necessary, since a very      *|
|*  similar function appears in utils.c, but we include it here so that    *|
|*  the hhelp.c module is entirely self-supporting, and relies on no       *|
|*  outside help at all.  It takes two char* arguments, template and	   *|
|*  try.  It returns true if the match, case-INsensitively, up to the	   *|
|*  end of try, provided try is non-null.				   *|
|*                                                                         *|
\***************************************************************************/

int hh_samestring (template, try)
  char *template;
  char *try;
{
  char *cp;
  
  if (try == NULL || *try == '\0')
    return (0);

  for (cp = try; *cp != 0; cp++)
    if (!((*cp == *(cp-try+template)) ||
	  ((islower (*cp)) && (toupper (*cp)) == *(cp-try+template)) ||
	  ((isupper (*cp)) && (tolower (*cp)) == *(cp-try+template))))
      return (0);
  
  return (1);
}

/***************************************************************************\
|*                                                                         *|
|*  The same comments apply here as to hh_samestring; the function         *|
|*  form() is one of the neat ideas to be whipped from C++ and is a kinda  *|
|*  sprintf() with its own static storage, so we don't have to keep doing  *|
|*  various static string definitons every time we wanna use a formatted   *|
|*  print/copy command, and then do sprintf into that buffer.		   *|
|*                                                                         *|
\***************************************************************************/

/*VARARGS*/
char *hh_form (va_alist)
  va_dcl
{
  va_list pvar;
  char *fmt_string;
  static char result[HH_LINELEN];
  
  va_start (pvar);
  fmt_string = va_arg (pvar, char*);
  (void) vsprintf (result, fmt_string, pvar);
  va_end (pvar);
  return (result);
}

/***************************************************************************\
|*                                                                         *|
|*  The function init_hhelp takes a pathname as its argument, and sets     *|
|*  up the global(ish) hh_root string to be this string; it should	   *|
|*  be the UNIX pathname to the root directory of the help heirachy.       *|
|*                                                                         *|
\***************************************************************************/

void init_hhelp (pathname)
  char *pathname;
{
  (void) strcpy (hh_root, pathname);
  (void) strcpy (hh_current, hh_root);
}

/***************************************************************************\
|*                                                                         *|
|*  The function hh_direntadd takes as its argument a pointer to a         *|
|*  directory entry (ie. a struct dirent), and adds its filename as a      *|
|*  new entry to the dynamically-allocated global character array,	   *|
|*  hh_files[][], containing the filename of the directory entry pointed   *|
|*  to by its parameter.  The return value of the {m,re}alloc() used for   *|
|*  this is ignored, since if it fails, we can do nothing about it	   *|
|*  anyway; as it is, if {m,re}alloc fails, we simply end up with an	   *|
|*  empty list of filenames.					       	   *|
|*                                                                         *|
\***************************************************************************/

void hh_direntadd (d_ptr)
  struct dirent *d_ptr;
{
  hh_nofiles++;
  if (hh_files == NULL)
    hh_files = (char**) malloc (sizeof (char*));
  else
    hh_files = (char**) realloc ((char*) hh_files,
				 (unsigned) hh_nofiles*sizeof (char*));

  hh_files[hh_nofiles-1] = strdup (d_ptr->d_name);
}

/***************************************************************************\
|*                                                                         *|
|*  The function make_list() constructs a list of the currently available  *|
|*  folders/texts in the global hh_files[][].			       	   *|
|*                                                                         *|
\***************************************************************************/

char *make_list ()
{
  DIR *dp;
  struct dirent *entry_ptr;
  
  if (hh_files != NULL)
    free ((char*) hh_files);	/* Zero the array of current filenames */
  hh_files = NULL;
  hh_nofiles = 0;

  if ((dp = opendir (hh_current)) == NULL)
    return;
  
  while ((entry_ptr = readdir (dp)) != NULL)
    if ((*entry_ptr->d_name != '.') &&
	(entry_ptr->d_name[strlen (entry_ptr->d_name)-1] != '~'))
      hh_direntadd (entry_ptr);
  
  (void) closedir (dp);
}

/***************************************************************************\
|*                                                                         *|
|*  The function hh_direntstr takes as its argument a pointer to a	   *|
|*  directory entry (ie. a struct dirent), and returns a string		   *|
|*  containing a text description of the directory entry.		   *|
|*                                                                         *|
|*  Format of each entry is one of the following:			   *|
|*   -- 2000000000 bytes: "index"					   *|
|*   -- [Folder]          "test_dir"					   *|
|*                                                                         *|
\***************************************************************************/

char *hh_direntstr (d_ptr)
  struct dirent *d_ptr;
{
  struct stat buffer;

  if (stat (hh_form ("%s/%s", hh_current, d_ptr->d_name), &buffer) == -1)
    return (hh_form (" --         [Unknown] \"%s\"\n", d_ptr->d_name));
  
  if (buffer.st_mode & S_IFDIR)
    return (hh_form (" --          [Folder] \"%s\"\n", d_ptr->d_name));
  
  if (!(buffer.st_mode & S_IFREG))
    return (hh_form (" --           [Weird] \"%s\"\n", d_ptr->d_name));
  
  return (hh_form (" -- %10ld bytes: \"%s\"\n", buffer.st_size,
		   d_ptr->d_name));
}

/***************************************************************************\
|*                                                                         *|
|*  The function list_dir() takes as its argument a string, the name of    *|
|*  a sub-directory of the directory hh_current, and returns a string	   *|
|*  containing a list of its contents.					   *|
|*                                                                         *|
\***************************************************************************/

char *list_dir (dir_name)
  char *dir_name;
{
  DIR *dp;
  struct dirent *entry_ptr;
  static char *list = NULL;
  char *cp;
  int nofiles = 0;

  if (list != NULL)
    free (list);

  if ((list = malloc (HH_LINELEN)) == NULL)
    return ("Couldn't allocate memory for directory listing.\n");
  
  if (strcmp (dir_name, "."))
    (void) sprintf (list, "Listing for Subtopic: \"%s\":\n", dir_name);
  else
    if (!strcmp (hh_current, hh_root))
      (void) strcpy (list, "Listing for Root Help Menu:\n");
    else
      if ((cp = rindex (hh_current, '/')) == NULL)
	(void) sprintf (list, "Current topic has a stupid pathname.\n");
      else
	(void) sprintf (list, "Listing for Subtopic: \"%s\":\n", cp+1);
  
  if ((dp = opendir (hh_current)) == NULL)
    return (hh_form ("Couldn't list contents of \"%s\".\n", dir_name));
  
  while ((entry_ptr = readdir (dp)) != NULL)
    if ((*entry_ptr->d_name != '.') &&
	(entry_ptr->d_name[strlen (entry_ptr->d_name)-1] != '~')) {
      nofiles++;
      cp = hh_direntstr (entry_ptr);
      if ((list = realloc (list, (unsigned) (strlen (list)+strlen (cp))))
	  == NULL)
	return ("Couldn't allocate memory for directory listing.\n");
      (void) strcat (list, cp);
    }
  
  (void) closedir (dp);
  
  if (nofiles == 0)
    return (hh_form ("%s -- No help in this subtopic.\n", list));
  else
    return (list);
}

/***************************************************************************\
|*                                                                         *|
|*  The function list_file() returns the contents of the named file.       *|
|*                                                                         *|
\***************************************************************************/

char *list_file (line, buffer)
  char *line;
  struct stat buffer;
{
  static char *cp = NULL;
  int fd;
  
  if (cp != NULL)		/* Ensure that cp is never multiply */
    free (cp);			/* allocated, and thus wasting memory. */
  
  if ((cp = malloc ((unsigned) buffer.st_size+1)) == NULL)
    return (hh_form ("Can't allocate memory for help on \"%s\".\n", line));
  
  if ((fd = open (hh_form ("%s/%s", hh_current, line), O_RDONLY)) == -1)
    return (hh_form ("Can't open file for help on \"%s\".\n", line));
  
  if (read (fd, cp, (int) buffer.st_size) != buffer.st_size)
    return (hh_form ("Can't read help on \"%s\".\n", line));

  cp[buffer.st_size] = '\0';
  return (cp);
}

/***************************************************************************\
|*                                                                         *|
|*  The function handle_hhline takes a line of text, which should be the   *|
|*  name of a file of help text, relative to the current hhelp-hirectory.  *|
|*  if it is the name of a directory, it returns a listing of the files    *|
|*  in it, (ie. hhelp subtopics).  In this case, the global array of	   *|
|*  characters hh_current[] is modifed to point at this new directory.	   *|
|*  (exception: if the directory-name is ".", we don't do this.  If the    *|
|*  file named is a plain file or a symbolic link, its contents will be    *|
|*  returned in a massive string.  If the file doesn't exist, a message    *|
|*  to that effect will be returned.					   *|
|*                                                                         *|
\***************************************************************************/

char *handle_hhfilename (line)
  char *line;
{
  struct stat buffer;

  if (stat (hh_form ("%s/%s", hh_current, line), &buffer) == -1)
    return (hh_form ("Couldn't find help on \"%s\".\n", line));

  if (buffer.st_mode & S_IFDIR) {
    if (strcmp (line, "."))
      (void) strcat (hh_current, hh_form ("/%s", line));

    make_list ();
    return (list_dir (line));
  }

  if (buffer.st_mode & S_IFREG)
    return (list_file (line, buffer));

  return (hh_form ("The file \"%s\" is too weird to list, sorry.\n", line));
}

/***************************************************************************\
|*                                                                         *|
|*  The function handle_hhline() determines to what filename the user is   *|
|*  attempting to refer, and calls handle_hhfilename() with that as its    *|
|*  argument.  The first line of attack is to see if the supplied line	   *|
|*  matches (in the sense of hh_samestring()) any of the stored filenames  *|
|*  in the global hh_files[][].  If not, then we attempt to read the file  *|
|*  or directory exactly as specified instead, (thus allowing queries to   *|
|*  probe directly through levels)					   *|
|*                                                                         *|
\***************************************************************************/

char *handle_hhline (line)
  char *line;
{
  int i;

  if (hh_files != NULL)
    for (i = 0; ((hh_files[i] != NULL) && (i < hh_nofiles)); i++)
      if (hh_samestring (hh_files[i], line))
	return (handle_hhfilename (hh_files[i]));

  return (handle_hhfilename (line));
}

/***************************************************************************\
|*                                                                         *|
|*  The function handle_hhelp() is the only function that should be        *|
|*  visible to the ouside world, other than the initialisation function    *|
|*  which I have just thought of, as a means for making the whole of the   *|
|*  hhelp(tm :-) package kinda integral and integrated and homogenous.     *|
|*  anyway: handle_hhelp() is the interface with the outside world.        *|
|*                                                                         *|
|*  Anyway, here comes the formal definition: the parameter line of        *|
|*  this function contains a command typed at the hhelp interpreter.       *|
|*  can be the name of a help file to list, a sub-directory to enter,      *|
|*  or an actual command as such ...  The only ones that I think will      *|
|*  exist are "break", (or "") to break a level, (ie change back to        *|
|*  the parent directory), and "quit" to exit completely.  More might      *|
|*  follow, who knows?  The function's return value is either a pointer    *|
|*  to a string to be printed, or otherwise handled, (ie. the text of      *|
|*  the help file, in one huge char* buffer), or NULL if the "quit"        *|
|*  command was sent.  (Whoops, wasn't quite as formal as I had meant      *|
|*  it to be, sorry)							   *|
|*                                                                         *|
|*  Oh look, a new command - "suspend" breaks out of hhelp mode, but	   *|
|*  in such a way that the current help directory remains when hhelp	   *|
|*  is next invoked, whereas if a hhelp session is terminated by "quit",   *|
|*  the directory is reset to the root (as set in init_hhelp()).  In	   *|
|*  either of these cases, handle_hhelp() returns NULL.			   *|
|*                                                                         *|
|*  Cheese, I've got another new command, and I think I'm gonna have to    *|
|*  write all this up properly later on.  Never mind.  The new command     *|
|*  is "list" and it re-lists the contents of the current help-directory   *|
|*  thus being semantically equivalent to ".".				   *|
|*                                                                         *|
|*  Frogging frogs, I've done it again!  The next new command is "where",  *|
|*  and it returns the current value of the global string hh_current, ie.  *|
|*  the pathname of the current help directory.				   *|
|*                                                                         *|
|*  The huge logical construction near the end of the function ensures     *|
|*  that any command that has a path-component on ".." in it will be       *|
|*  rejected, (otherwise it would be a security risk to use the hhelp      *|
|*  system in a set[ug]id program, since it would be impossible to limit   *|
|*  the user to reading files in a specified directory heirachy).  The     *|
|*  arrangement used here is like chroot()ing to hh_root.		   *|
|*                                                                         *|
|*  Yet another new command rears its ugly head - "debug" will return      *|
|*  whatever debugging infomation I happen to code into it; to start       *|
|*  with, I think the contents of hh_files will do nicely.		   *|
|*                                                                         *|
|*  Well, since there are now an infinite number of commands anyway,	   *|
|*  I'm gonna add another one, "help", which will list all the commands.   *|
|*                                                                         *|
\***************************************************************************/

char *handle_hhelp (line)
  char *line;
{
  char *cp, *frog = NULL;
  int i;

  if (hh_samestring ("help", line)) {
    return ("\n\
Heirachical Help Commands\n\
=========================\n\n\
<text-name>	Print the named text.\n\
<folder-name>	Change into the named folder.\n\
break		Return to the previous folder.\n\
list		List contents of current folder.\n\
where		Print pathname of current folder.\n\
suspend		Suspend current hhelp session.\n\
quit		Quit current hhelp session.\n\
debug		Print debugging information.\n\
	(Fuller instructions can be found in the folder hhelp,\n\
	 in the category \"using_the_system\")\n\n");
  }

  if ((hh_samestring ("break", line)) || (!strcmp (line, ".."))) {
    if (!strcmp (hh_root, hh_current))
      return ("Already at top level.\n");
    
    if ((cp = rindex (hh_current, '/')) == NULL)
      return ("Pathname is too odd.  Go away.\n");

    *cp = '\0';
    make_list ();
    return (hh_form ("Broke a level: New folder is HH_ROOT%s\n",
		     hh_current+strlen (hh_root)));
  }
  
  if ((hh_samestring ("list", line)) || (!strcmp (line, "."))) {
    return (handle_hhline ("."));
  }
  
  if (hh_samestring ("where", line)) {
    return (hh_form ("Current folder is HH_ROOT%s\n",
		     hh_current+strlen (hh_root)));
  }
  
  if (hh_samestring ("suspend", line)) {
    return (NULL);
  }
  
  if (hh_samestring ("quit", line)) {
    (void) strcpy (hh_current, hh_root);
    return (NULL);
  }
  
  if (hh_samestring ("debug", line)) {
    if (hh_files == NULL)
      return (hh_form ("No filenames stored (should be %d)\n", hh_nofiles));

    for (i = 0; ((hh_files[i] != NULL) && (i < hh_nofiles)); i++) {
      cp = hh_form (" -- \"%s\"\n", hh_files[i]);

      if (frog == NULL) {
	if ((frog = malloc ((unsigned) (strlen (cp)+1))) == NULL)
	  return ("Couldn't allocate debugging memory.\n");
	(void) strcpy (frog, cp);
      }

      else {
	if ((frog = realloc (frog, (unsigned) (strlen (frog)+strlen (cp)+1)))
	    == NULL)
	  return ("Couldn't allocate enough debugging memory.\n");
	(void) strcat (frog, cp);
      }
    }
    
    return (frog);
  }
  
  if (strlen (line) > 1)
    for (cp = line; cp[1] != '\0'; cp++)
      if (((cp == line) || (cp[-1] == '/')) &&
	  ((*cp == '.') && (cp[1] == '.')) &&
	  ((cp[2] == '/') || (cp[2] == '\0')))
	return ("Can't use \"..\" as a path component.\n");
  
  for (cp = line; *cp != '\0'; cp++)
    if (((cp == line) || (cp[-1] == '/')) &&
	(*cp == '.') &&
	((cp[1] == '/') || (cp[1] == '\0')))
      return ("Can't use \".\" as a path component.\n");
  
  return (handle_hhline (line));
}

/*-------------------------------------------------------------------------*/
